"""
Datetime utilities for converting to and from datetime.datetime, json, naive and time zone aware
"""

from datetime import datetime


from dateutil import tz
from timezonefinder import TimezoneFinder

from coordinate_projector.projector import Projector

projector = Projector()

_time_zone_finder: TimezoneFinder | None = None


def ensure_tz(
    dt: datetime | None,
    longitude: float | None = None,
    latitude: float | None = None,
    srid: int = 4326,
) -> datetime | None:
    """
    Return passed datetime dt enriched with timezone.

    If timezone already present in the input then it is returned untouched.

    If datetime does not contain a timezone, then try to find the timezone by the
    optional location (longitude, latitude in EPSG:4326 in degrees).

    If no location or timezone is provided, then assume the passed datetime is
    recorded in the norwegian timezone.
    """
    global _time_zone_finder

    if not dt:
        return dt

    if not isinstance(dt, datetime):
        raise Exception("Got unexpected type for datetime!")

    if dt.tzinfo is None or dt.tzinfo.utcoffset(dt) is None:
        # timezone naive (no time zone in dt)
        if longitude is not None and latitude is not None:
            if srid and srid != 4326:
                longitude, latitude = projector.transform(from_srid=srid, to_srid=4326, east=longitude, north=latitude)

            # find timezone from position
            if not _time_zone_finder:
                _time_zone_finder = TimezoneFinder()

            input_timezone = tz.gettz(_time_zone_finder.timezone_at(lng=longitude, lat=latitude))
        else:
            # Assume Norway
            input_timezone = tz.gettz("Europe/Oslo")

        dt = dt.replace(tzinfo=input_timezone)

    return dt


def datetime_to_json(
    dt: datetime | None,
    longitude: float | None = None,
    latitude: float | None = None,
    srid: int = 4326,
) -> str | None:
    """
    Return passed datetime.datetime as json-formatted (iso-8601) string with UTC timezone. Sub-second time information
    is removed.

    If datetime contains timezone information that timezone is used.
    If datetime does not contain a timezone, then assume timezone by the
    optional recording location (longitude, latitude in EPSG:4326 in degrees).
    If no location or timezone is provided, then assume the passed datetime is recorded in the norwegian timezone.
    """
    if not dt:
        return None

    dt = ensure_tz(dt, longitude, latitude, srid=srid)

    if not dt:
        return None

    return dt.replace(microsecond=0).astimezone(tz.UTC).isoformat()



from functools import cache

import pyproj
from pyproj import Transformer, CRS

_transformers: dict[str, Transformer] = {}


class Projector:
    @staticmethod
    @cache
    def get_supported_projections() -> dict[str, CRS]:
        all_crs_codes = pyproj.database.get_codes("EPSG", pyproj.enums.PJType.PROJECTED_CRS, False)

        return {code: pyproj.CRS.from_epsg(code) for code in all_crs_codes} | {"4326": pyproj.CRS.from_epsg("4326")}

    @staticmethod
    def _get_transformer(from_srid: int | str, to_srid: int | str) -> Transformer:
        global _transformers

        if transformer := _transformers.get(f"{from_srid}-{to_srid}"):
            return transformer

        from_crs = CRS.from_epsg(from_srid)
        to_crs = CRS.from_epsg(to_srid)
        transformer = Transformer.from_crs(from_crs, to_crs, always_xy=True)

        _transformers[f"{from_srid}-{to_srid}"] = transformer

        return transformer

    def transform(self, from_srid: int | str, to_srid: int | str, east: float, north: float) -> tuple[float, float]:
        transformer: Transformer = self._get_transformer(from_srid, to_srid)
        projected_east, projected_north = transformer.transform(east, north)
        return projected_east, projected_north
